# 20.7 使用数据存储

我们现在有了一种用 html 窗口来收集用户信息的方法。通常我们希望使这些信息持久化：我们需要一个地方来放置这些信息，并且需要一个方法来取回这些信息。GAE 在这里为我们提供了它的 DataStore 设施：一个非关系型 (non-relational) 数据库，它可以跨网络服务器甚至跨机器持久保存您的数据。事实上，用户的下一个请求很可能是在不同的计算机上运行的不同的网络服务器，但 GAE 的基础设施在一个简单的 API 后面处理了所有的数据分布、复制和负载平衡，你还可以得到一个强大的查询引擎。

我们现在将对我们的例子进行一些扩展，制作一个问候结构，该结构可以包含问候的作者、内容和时间，我们要存储这些内容。这是你要做的第一件事：为你的程序*实体*（即你的程序所处理的对象）制作一个合适的数据结构，大多数情况下这将是一个 `struct`。在运行的程序中，这个结构的内存值将包含来自该实体的 DataStore 的数据。

接下来我们程序的版本如下：

​		(A) `url:/`： 检索所有存储的问候语并通过 `template` 包显示它们

​		(B) `url:/sign`：存储一个新的问候语到数据存储里面

我们现在需要导入 `appengin/datastore` 包：

<u>[Listing 20.5 helloworld2_version4.go:](examples\chapter_20\helloapp\hello\helloworld2_version4.go)</u>

```go
package hello

import (
	"appengine"
	"appengine/datastore"
	"appengine/user"
	"net/http"
	"template"
	"time"
)

const guestbookTemplateHTML = `
<html>
  <body>
    {{range .}}
      {{with .Author}}
        <p><b>{{html .}}</b> wrote:</p>
      {{else}}
        <p>An anonymous person wrote:</p>
      {{end}}
      <pre>{{html .Content}}</pre>
      <pre>{{html .Date}}</pre>
    {{end}}
    <form action="/sign" method="post">
      <div><textarea name="content" rows="3" cols="60"></textarea></div>
      <div><input type="submit" value="Sign Guestbook"></div>
    </form>
  </body>
</html>
`

var guestbookTemplate = template.Must(template.New("book").Parse(guestbookTemplateHTML))

type Greeting struct {
	Author  string
	Content string
	Date    datastore.Time
}

func init() {
	http.HandleFunc("/", root)
	http.HandleFunc("/sign", sign)
}

func root(w http.ResponseWriter, r *http.Request) {
	c := appengine.NewContext(r)
	q := datastore.NewQuery("Greeting").Order("-Date").Limit(10)
	greetings := make([]Greeting, 0, 10)
	if _, err := q.GetAll(c, &greetings); err != nil {
		http.Error(w, err.String(), http.StatusInternalServerError)
		return
	}
	if err := guestbookTemplate.Execute(w, greetings); err != nil {
		http.Error(w, err.String(), http.StatusInternalServerError)
	}
}

func sign(w http.ResponseWriter, r *http.Request) {
	c := appengine.NewContext(r)
	g := Greeting{
		Content: r.FormValue("content"),
		Date:    datastore.SecondsToTime(time.Seconds()),
	}
	if u := user.Current(c); u != nil {
		g.Author = u.String()
	}
	_, err := datastore.Put(c, datastore.NewIncompleteKey(c, "Greeting", nil), &g)
	if err != nil {
		http.Error(w, err.String(), http.StatusInternalServerError)
		return
	}
	http.Redirect(w, r, "/", http.StatusFound)
}
```

信号处理程序 (B) 从表单和上下文数据中构建了一个问候值 `g`，然后用 `datastore.Put()` 存储它。DataStore 在内部为数据记录生成自己的唯一 key；为了做到这一点，我们调用 `Put()` 函数，将

```go
datastore.NewIncompleteKey(c, "Greeting", nil)
```

作为第 2 个参数（这个函数需要实体 `Greeting` 的名字）。`Put()`  的第 3 个参数 `&g` 是一个包含 value 的 `struct` （严格来说，此处应为指向这个 `struct` 的指针）。

`datastore` 包提供了一个查询类型，用于查询数据存储并迭代结果。根处理程序正是通过构造一个查询 `q()` 来实现的，该查询按照日期降序从DataStore 中请求问候对象，限制为 10 个。

```go
q := datastore.NewQuery("Greeting").Order("-Date").Limit(10)
```

我们需要一个数据结构来存储我们的查询结果，也就是 `greetings`，一个 Greeting 值的切片。我们对 `q.GetAll(c, &greetings)` 的调用检索了数据，并将它们存储在我们的切片中；当然，我们会检查可能的错误。

当一切正常时，我们通过与我们的模板合并来显示数据：

```go
guestbookTemplate.Execute(w, greetings)
```

这是由一个 `range` 结构执行的（参考 [15.7.6 节](15.7.md)）。

再次通过编辑 helloworld2.go 文件进行测试，用 listing 20.5 中的代码替换它；在问候中间的间隙关闭浏览器会话，这样你可以检查它们是否被持久保存。

<u>清除开发用服务器的数据存储：</u>

开发用 web 服务器使用一个本地版本的数据存储来测试你的应用程序，使用临时文件。只要临时文件存在，数据就会持续存在，除非你要求，否则 Web 服务器不会重置这些文件。如果你想让开发用服务器在启动前擦除其数据存储，请在启动服务器时使用 `--clear_datastore` 选项：

```bash
dev_appserver.py --clear_datastore helloapp/
```

<u>调试：</u>

gdb 调试器可以和 Go 一起使用（见 http://golang.org/doc/debugging_with_gdb.html），你可以将 gdb 附加到一个现有的进程中。因此：像往常一样启动 dev_appserver.py，并访问 `localhost:8080` 来启动你的 Go 应用程序。然后执行： `$ ps ax | grep _go_app`，找到 \_go\_app 的 PID 和路径。如果你把 gdb 连接到这个上面，那么你对 dev_appserver 的下一个 HTTP 请求应该会碰到你在代码中设置的任何断点。记住，如果你修改了 Go 的源代码，那么开发应用服务器将重新编译并执行不同的 \_go\_app。

## 链接

- [目录](directory.md)
- 上一节：[处理窗口](20.6.md)
- 下一节：[上传到云端](20.8.md)

